@@ -273,7 +273,8 @@ class Converter():
                     if copy_lights:
                         light = light.make_copy()
                     lnp = np.attach_new_node(light)
-                    lnp.set_p(lnp.get_p() - 90)
+                    if self.compose_cs == CS_zup_right:
+                        lnp.set_p(lnp.get_p() - 90)
                     lnp.set_r(lnp.get_r() - 90)
                     if isinstance(light, Light):
                         root.set_light(lnp)
@@ -466,8 +467,8 @@ class Converter():
         self.textures['__pbr-fallback'] = texture
 
         texture = Texture('emission-fallback')
-        texture.setup_2d_texture(1, 1, Texture.T_unsigned_byte, Texture.F_rgba)
-        texture.set_clear_color(LColor(0, 0, 0, 0))
+        texture.setup_2d_texture(1, 1, Texture.T_unsigned_byte, Texture.F_luminance)
+        texture.set_clear_color(LColor(1, 1, 1, 1))
 
         self.textures['__emission-fallback'] = texture
 
@@ -645,9 +646,16 @@ class Converter():
             texinfos[-1]['mode'] = TextureStage.M_normal
 
             # Emission map
-            pmat.set_emission(LColor(*gltf_mat.get('emissiveFactor', [0.0, 0.0, 0.0]), w=0.0))
-            texinfos.append(gltf_mat.get('emissiveTexture', emission_fallback))
-            texinfos[-1]['mode'] = TextureStage.M_add
+            factor = gltf_mat.get('emissiveFactor', [0.0, 0.0, 0.0])
+            pmat.set_emission(LColor(*factor, w=0.0))
+            if 'emissiveTexture' in gltf_mat:
+                texinfos.append(gltf_mat['emissiveTexture'])
+                texinfos[-1]['mode'] = TextureStage.M_add
+            else:
+                # The fallback texture is all-white, so set it to "modulate" to
+                # make sure the model isn't affected by it without simplepbr.
+                texinfos.append(emission_fallback)
+                texinfos[-1]['mode'] = TextureStage.M_modulate
             if texinfos[-1]['index'] in self.textures:
                 self.make_texture_srgb(self.textures[texinfos[-1]['index']])
 
@@ -661,6 +669,7 @@ class Converter():
 
         # Setup textures
         tex_attrib = TextureAttrib.make()
+        tex_mat_attrib = None
         for i, texinfo in enumerate(texinfos):
             texdata = self.textures.get(texinfo['index'], None)
             if texdata is None:
@@ -673,7 +682,40 @@ class Converter():
             texstage.set_mode(texinfo.get('mode', TextureStage.M_modulate))
             tex_attrib = tex_attrib.add_on_stage(texstage, texdata)
 
+            transform_ext = texinfo.get('extensions', {}).get('KHR_texture_transform')
+            if transform_ext:
+                if 'texCoord' in transform_ext:
+                    # This overrides, if present.
+                    texstage.set_texcoord_name(InternalName.get_texcoord_name(str(transform_ext['texCoord'])))
+
+                # glTF uses a transform origin of the upper-left corner of the
+                # texture, whereas Panda uses the lower-left corner.
+                mat = Mat3()
+                scale = transform_ext.get('scale')
+                if scale:
+                    mat *= (Mat3.translate_mat(0, -1) *
+                            Mat3.scale_mat(scale[0], scale[1]) *
+                            Mat3.translate_mat(0, 1))
+
+                rot = transform_ext.get('rotation')
+                if rot:
+                    mat *= (Mat3.translate_mat(0, -1) *
+                            Mat3.rotate_mat(math.degrees(rot)) *
+                            Mat3.translate_mat(0, 1))
+
+                offset = transform_ext.get('offset', [0, 0])
+                if offset:
+                    mat *= Mat3.translate_mat(offset[0], -offset[1])
+
+                transform = TransformState.make_mat3(mat)
+                if not tex_mat_attrib:
+                    tex_mat_attrib = TexMatrixAttrib.make(texstage, transform)
+                else:
+                    tex_mat_attrib = tex_mat_attrib.add_stage(texstage, transform)
+
         state = state.set_attrib(tex_attrib)
+        if tex_mat_attrib:
+            state = state.set_attrib(tex_mat_attrib)
 
         # Setup Alpha mode
         alpha_mode = gltf_mat.get('alphaMode', 'OPAQUE')
@@ -764,6 +806,7 @@ class Converter():
         accessors = sorted(accessors, key=lambda x: x['bufferView'])
         data_copies = []
         is_skinned = 'JOINTS_0' in mesh_attribs
+        calc_normals = not 'NORMAL' in mesh_attribs
         calc_tangents = not 'TANGENT' in mesh_attribs
 
         for buffview, accs in itertools.groupby(accessors, key=lambda x: x['bufferView']):
@@ -948,11 +991,43 @@ class Converter():
         #print(ss.data.decode('utf8'))
         geom = Geom(vdata)
         geom.add_primitive(prim)
+        if calc_normals:
+            self.calculate_normals(geom)
         if calc_tangents:
             self.calculate_tangents(geom)
         geom.transform_vertices(self.csxform)
         geom_node.add_geom(geom, mat)
 
+    def calculate_normals(self, geom):
+        # Generate flat normals, as required by the glTF spec.
+        if geom.get_primitive_type() != GeomEnums.PT_polygons:
+            return
+
+        # We need to deindex the primitive since each occurrence of a vertex on
+        # a triangle could have a different normal vector.
+        geom.decompose_in_place()
+        geom.make_nonindexed(False)
+
+        gvd = geom.get_vertex_data()
+        gvd = gvd.replace_column(InternalName.get_normal(), 3, GeomEnums.NT_float32, GeomEnums.C_normal)
+        vertex_reader = GeomVertexReader(gvd, 'vertex')
+        normal_writer = GeomVertexWriter(gvd, 'normal')
+
+        read_vertex = vertex_reader.get_data3
+        write_normal = normal_writer.set_data3
+
+        while not vertex_reader.is_at_end():
+            vtx1 = read_vertex()
+            vtx2 = read_vertex()
+            vtx3 = read_vertex()
+            normal = (vtx2 - vtx1).cross(vtx3 - vtx1)
+            normal.normalize()
+            write_normal(normal)
+            write_normal(normal)
+            write_normal(normal)
+
+        geom.set_vertex_data(gvd)
+
     def calculate_tangents(self, geom):
         # Adapted from https://www.marti.works/calculating-tangents-for-your-mesh/
         prim = geom.get_primitive(0)
@@ -1527,10 +1602,6 @@ class Converter():
 
         # Update the light
         if punctual:
-            coneangle = math.pi / 4
-            if 'spot' in gltf_light and 'outerConeAngle' in gltf_light['spot']:
-                coneangle = gltf_light['spot']['outerConeAngle']
-
             # For PBR, attention should always be (1, 0, 1)
             if hasattr(node, 'attenuation'):
                 node.attenuation = LVector3(1, 0, 1)
@@ -1540,8 +1611,20 @@ class Converter():
             if 'range' in gltf_light:
                 node.max_distance = gltf_light['range']
             if ltype == 'spot':
-                fov = math.degrees(coneangle) * 2
+                spot = gltf_light.get('spot', {})
+                inner = spot.get('innerConeAngle', 0)
+                outer = spot.get('outerConeAngle', math.pi / 4)
+                fov = math.degrees(outer) * 2
                 node.get_lens().set_fov(fov, fov)
+
+                if inner >= outer:
+                    node.exponent = 0
+                else:
+                    # The value of exp was chosen empirically to give a smooth
+                    # cutoff without straying too far from the spec; higher
+                    # exponents will have a smoother cutoff but sharper falloff.
+                    exp = 8 / 3
+                    node.exponent = 2 * (math.pi * 0.5 / outer) ** exp
         else:
             if ltype == 'unsupported':
                 lightprops = {}
